#!/usr/bin/env python
# -*- coding: utf-8 -*-

import redis
from config import Config


class Redis(object):

    """ Borg Pattern
    one state many objects useful in this case as we
    want one pool but many objects
    """
    __state = {}

    def __init__(self):
        self.__dict__ = self.__state
        if '_pool' not in self.__dict__:
            config = Config()
            # we only need one pool per app
            self._pool = redis.ConnectionPool(db=config.redis.get('db', 0),
                                              host=config.redis.get('host', 'localhost'),
                                              port=config.redis.get('port', 6379),
                                              max_connections=config.redis.get('max_poolsize', 10))

        # each object required its own connection so that we can close
        # it after each request without affecting other requests
        #
        # so this line is NOT within the above if statement
        self._redis = redis.StrictRedis(connection_pool=self._pool)

    def __getattr__(self, name):
        return getattr(self._redis, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import bottle


class BaseSingleton(type):

    """
    Need to write this because supports both Python 2.x and 3.x.

    class SubClass(with_metaclass('SubClass', Singleton)):
        pass
    """
    __instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls.__instances:
            cls.__instances[cls] = super(BaseSingleton, cls).__call__(*args, **kwargs)
        return cls.__instances[cls]

    def get_instance(cls):
        bottle.depr("Please use {0}() instead of {0}.get_instance()".format(cls.__name__))
        return cls.__call__()


# Hide how to write a complex inheritance
class Singleton(BaseSingleton('Singleton', (object, ), {})):

    """
    It becomes Singleton When class inherit this.
    """
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function
import os
import sys
import textwrap
import pwd
import grp
from gevent import monkey
monkey.patch_all()
from gevent import pywsgi
import bottle
from datetime import date
from vendor.daemon import Daemon
from lib.middleware import Dispatcher, StripPath, SessionWsgi
from lib.config import Config
import lib.plugin
from lib.logger import Log
import argparse

original_args = []


class Decanter(Daemon):

    def __init__(self, app, hostname='localhost', port=9000,
                 pidfile='/var/run/decanter.pid', development=False):
        self.app = app
        self.hostname = hostname
        self.port = int(port)
        self.pidfile = pidfile
        self.config = Config()

        if 'timezone' in self.config:
            os.environ['TZ'] = self.config.timezone

        # remove all default bottle plugins
        bottle.uninstall(True)
        bottle.DEBUG = self.config.debug
        # install plugins
        self.install(plugins=self.config.plugins)
        if self.config.debug or not development:
            stdout = os.popen('tty').read().strip()
            stderr = os.popen('tty').read().strip()

        if not development:
            super(Decanter, self).__init__(
                pidfile, stdout=stdout, stderr=stderr)

    def install(self, plugins=[]):
        third_party_plugin_dir = '/'.join([self.config.apppath, 'plugins'])
        third_party_plugin_module = '.'.join([
            self.config.apppath.strip(os.path.sep).split(os.path.sep).pop(),
            'plugins'])
        for plugin in plugins:
            name = ''.join([plugin.capitalize(), 'Plugin'])
            cls = getattr(lib.plugin, name, None)

            if not cls and os.path.isdir(third_party_plugin_dir):
                for plugin_file in os.listdir(third_party_plugin_dir):
                    if not plugin_file.endswith('.py'):
                        continue
                    module = __import__(
                        '.'.join([
                            third_party_plugin_module,
                            os.path.splitext(plugin_file)[0]]),
                        fromlist=[name])
                    cls = getattr(module, name, None)
                    if cls:
                        break

            if not cls:
                raise ImportWarning(name + ' is not found.')
            bottle.install(cls())

    def daemonize(self):
        haspid = os.path.isfile(self.pidfile)

        super(Decanter, self).daemonize()

        username = pwd.getpwuid(os.getuid()).pw_name
        if username != self.config.user:
            uid = pwd.getpwnam(self.config.user).pw_uid
            gid = grp.getgrnam(self.config.group).gr_gid
            os.chown(self.pidfile, uid, gid)
            os.setgid(gid)
            os.setuid(uid)

        if not haspid and os.path.isfile(self.pidfile)\
                and not self.config.test:
            print("Starting daemon with pidfile: {0}".format(self.pidfile))

    def run(self):
        log = None if self.config.test else 'default'
        server = pywsgi.WSGIServer((
            self.hostname, self.port), self.app, log=log)
        server.serve_forever()

    def runserver(self):
        """
        Command to run for development environments.
        """
        print(textwrap.dedent("""
            {2}
            Welcome to Decanter!
            You're ready to go!
            Server running on http://{1}:{0}
            {2}""".format(self.port, self.hostname, '=' * 50)))
        try:
            self.run()
        except KeyboardInterrupt:
            # Don't show traceback when sending ^C during runserver
            print('\nGood-bye!')

    def status(self):
        try:
            with open(self.pidfile, 'r') as fp:
                pid = int(fp.read())
            os.kill(pid, 0)
            print("Decanter is running, pidfile: {0}, process: {1}".format(
                self.pidfile, pid))
        except (OSError, IOError):
            print("Decanter is not running")


def parse_args(filepath=__file__, source=sys.argv, custom_commands=[]):
    """
    This function will parse command line arguments. To display help and exit
    if the argument is invalid. Will return command, hostname, port and config.
    """
    global original_args
    original_args = sys.argv[:]

    defaults = {
        'myself': source.pop(0),
        'hostname': 'localhost',
        'port': 9000,
        'conf': 'app/config/settings.py'
    }
    if len(source) == 0:
        source.append(defaults['myself'])

    parser = argparse.ArgumentParser(
        description='Example: {myself} -h {hostname}' +
                    '-p {port} -c config/devel.py start'.format(
                        **defaults), conflict_handler='resolve')
    parser.add_argument('command', choices=[
                        'start', 'stop', 'restart', 'status', 'runserver'] + custom_commands)
    parser.add_argument('-h', '--hostname', default=defaults['hostname'])
    parser.add_argument('-p', '--port', type=int, default=defaults['port'])
    parser.add_argument(
        '-c', '--config', required=False, default=defaults['conf'], type=argparse.FileType(),
        help='config must match the location of a module containing' +
             'decanter required configuration items, i.e. app/config/settings.py')
    args = parser.parse_args(source)

    # 'type=argparse.FileType()' will confirm the existence of a file.
    # but it open file.
    args.config.close()
    args.config = os.path.relpath(os.path.realpath(args.config.name),
                                  os.path.dirname(os.path.realpath(filepath)))

    return args

if __name__ == '__main__':
    args = parse_args()

    # initialize Config
    config = Config(args.config)

    # the log file
    logfile = config.logger['filepath'].format(args.port, date.today())
    # the pid fle
    pidfile = config.pidfile.format(args.port)
    # initialize logger
    log = Log(logfile)

    # make directory to put pid file
    piddir = os.path.dirname(pidfile)
    if not os.path.isdir(piddir):
        os.makedirs(piddir)

    # the app
    app = Dispatcher(StripPath(SessionWsgi(bottle.app())), config)
    decanter = Decanter(
        app, hostname=args.hostname, port=args.port, pidfile=pidfile,
        development=args.command == 'runserver')

    # execute command
    {
        'start': lambda: decanter.start(),
        'stop': lambda: decanter.stop(),
        'status': lambda: decanter.status(),
        'restart': lambda: decanter.restart(),
        'runserver': lambda: decanter.runserver(),
    }[args.command]()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re
import json
import gettext
import traceback
from functools import wraps
from jinja2 import BaseLoader
from jinja2 import Environment
from jinja2 import ChoiceLoader
from jinja2 import FileSystemLoader
from jinja2 import TemplateNotFound
from bottle import request, response, PluginError
from .logger import Log
from .config import Config
from .errors import BaseError, ValidationError, ConnectionError

# Format of http.request.header.Accept-Language.
# refs: http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.4
REQUEST_ACCEPT_LANGUAGE_RE = re.compile(r'''
        ([A-Za-z]{1,8}(?:-[A-Za-z]{1,8})*|\*)   # "en", "en-au", "*"
        (?:;q=(0(?:\.\d{,3})?|1(?:.0{,3})?))?   # Optional "q=1.00", "q=0.8"
        (?:\s*,\s*|$)                           # Multiple accepts per header.
        ''', re.VERBOSE)


class DecanterLoader(BaseLoader):

    def __init__(self, path):
        self.path = path

    def get_source(self, environment, template):
        template = template.lstrip('/')
        bundle = template.partition('/')
        filepath = os.path.join(self.path, bundle[0], 'views', bundle[2])
        if os.path.isabs(filepath) and os.path.isfile(filepath):
            mtime = os.path.getmtime(filepath)
            with file(filepath) as f:
                source = f.read().decode('utf-8')
            return (source,
                    filepath,
                    lambda: mtime == os.path.getmtime(filepath))
        raise TemplateNotFound(template)


class Jinja2Plugin(object):
    __state = {}
    name = 'jinja2'
    api = 2

    def __init__(self):
        """ Borg Pattern """
        self.__dict__ = self.__state
        if 'env' not in self.__dict__:
            self.config = Config()
            basepath = os.path.join(self.config.apppath, 'bundles')
            bundels = [name for name in os.listdir(
                basepath) if os.path.isdir(os.path.join(basepath, name))]
            views = []

            for bundle in bundels:
                if bundle != 'views':
                    baseViews = os.path.join(basepath, bundle, 'views')
                    if os.path.isdir(baseViews):
                        views.append(baseViews)

            # add views directory in bundle
            if 'views' in bundels:
                views.append(os.path.join(basepath, 'views'))

            # add app/views directory
            views.append(os.path.join(self.config.apppath, 'views'))

            self.env = Environment(
                loader=ChoiceLoader([FileSystemLoader(views, encoding='utf-8'),
                                     DecanterLoader(basepath)]))

    def apply(self, callback, route):
        @wraps(callback)
        def wrapper(*args, **kwargs):
            if self.name in route.skiplist:
                return callback(*args, **kwargs)

            data = callback(*args, **kwargs)
            if not isinstance(data, dict):
                return data

            bundle = callback.__module__.split('.')[-3]
            controller = callback.__module__.split('.')[-1]
            action = callback.__name__
            template = os.path.join(
                bundle, controller, '.'.join([action, 'html']))
            try:
                tpl = self.env.get_template(template)
            except TemplateNotFound:
                template = os.path.join(controller, '.'.join([action, 'html']))
                tpl = self.env.get_template(template)
            return tpl.render(data)

        return wrapper

    def setup(self, app):
        pass

    def close(self):
        pass


class Jinja2i18nPlugin(Jinja2Plugin):
    __state = {}
    name = 'jinja2i18n'
    api = 2

    def __init__(self):
        super(Jinja2i18nPlugin, self).__init__()

        config = Config()
        self.domain = getattr(config, 'domain', 'locale')
        locale_dir = getattr(config, 'locale_dir')
        lang_code = getattr(config, 'lang_code', 'en')

        if locale_dir is None:
            locale_dir = os.path.normpath(os.path.join(
                os.path.dirname(os.path.abspath(__file__)), 'locale'))
        if not os.path.exists(locale_dir):
            raise PluginError(
                'No locale directory found, please assign a right one.')
        self.locale_dir = locale_dir
        self.lang_code = lang_code
        self.prepared = {}
        self.app = None
        self.trans = None

    def setup(self, app):
        self.app = app
        self.app._ = lambda s: s
        self.app.set_lang = self.prepare
        self.app.hooks.add('before_request', self.prepare)

    def extra_client_expected_langs(self):
        """Return language list from http.request.header.Accept-Language, ordered by 'q'."""
        result = []
        pieces = REQUEST_ACCEPT_LANGUAGE_RE.split(
            request.headers.get('Accept-Language', ''))
        if pieces[-1]:
            return []
        for i in range(0, len(pieces) - 1, 3):
            first, lang, priority = pieces[i: i + 3]
            if lang == '*':  # TODO: support default language
                return []
            if first:
                return []
            priority = priority and float(priority) or 1.0
            result.append((lang, priority))
        result.sort(key=lambda k: k[1], reverse=True)
        return result

    def get_language_list(self):
        expected_langs = self.extra_client_expected_langs()

        lang_codes = []

        for lang, priority in expected_langs:
            lang_country = lang.split('-')
            if len(lang_country) == 1:
                lang_codes.append(lang)
                continue
            country = lang_country[1]
            lang_codes.append('%s_%s' % (lang_country[0], country))
            lang_codes.append('%s_%s' % (lang_country[0], country.swapcase()))

        if not self.lang_code is None:
            lang_codes += [self.lang_code]

        return lang_codes

    def prepare(self, langs=None):
        if langs is None:
            langs = self.get_language_list()
        prepared_key = tuple(langs)
        if prepared_key in self.prepared:
            trans = self.prepared.get(prepared_key)
            if trans:
                trans.install(True)
                self.app._ = trans.ugettext
            else:
                self.app._ = lambda s: s

            self.trans = trans
            return

        try:
            trans = gettext.translation(
                self.domain, self.locale_dir, languages=langs)
            trans.install(True)
            self.app._ = trans.ugettext
            self.prepared[prepared_key] = trans
        except Exception:
            Log.get_instance().error("No locale folder found for translations or problems with importing it.")
            trans = None
            self.app._ = lambda s: s
            self.prepared[prepared_key] = None

        self.trans = trans

    def apply(self, callback, route):
        @wraps(callback)
        def wrapper(*args, **kwargs):
            if self.name in route.skiplist:
                return callback(*args, **kwargs)

            self.prepare()

            data = callback(*args, **kwargs)
            if not isinstance(data, dict):
                return data

            bundle = callback.__module__.split('.')[-3]
            controller = callback.__module__.split('.')[-1]
            action = callback.__name__
            template = os.path.join(
                bundle, controller, '.'.join([action, 'html']))
            try:
                tpl = self.env.get_template(template)
            except TemplateNotFound:
                template = os.path.join(controller, '.'.join([action, 'html']))
                tpl = self.env.get_template(template)

            # add i18n variables
            data = callback(*args, **kwargs)
            data['_'] = self.app._
            data['ngettext'] = self.trans.ungettext

            return tpl.render(data)

        return wrapper


class JsonPlugin(object):
    name = 'json'
    api = 2

    def apply(self, callback, route):
        @wraps(callback)
        def wrapper(*args, **kwargs):
            config = Config()
            if self.name in route.skiplist:
                return callback(*args, **kwargs)

            try:
                data = callback(*args, **kwargs)
            # catch validation errors from the controller
            except (BaseError, ValidationError, ConnectionError, Exception) as e:
                # create a standardized error object
                data = {
                    'opstat': 'error'
                }
                if hasattr(e, 'message') and e.message:
                    data['error'] = e.message
                elif hasattr(e, 'msg') and e.msg:
                    data['error'] = e.msg
                else:
                    data['error'] = unicode(e)
                if hasattr(e, 'fields') and isinstance(e.fields, dict):
                    data['fields'] = e.fields
                if hasattr(e, 'returned') and isinstance(e.returned, dict):
                    data['response'] = e.returned

                if not isinstance(e, ValidationError):
                    Log.get_instance().error(
                        "Error tracked: \Request: %s\nMessage: %s\nFields: %s\nResponse: %s\n\nTraceback: %s" % (
                            request,
                            e.message,
                            getattr(e, 'fields', None),
                            getattr(e, 'response', None),
                            traceback.format_exc()
                        )
                    )
                    # return server error response
                    response.status = 500

                if config.debug is True:
                    data['traceback'] = traceback.format_exc()

            response.set_header('Content-Type', 'application/json')
            return json.dumps(data)

        return wrapper

    def setup(self, app):
        pass

    def close(self):
        pass


class SessionPlugin(object):
    name = 'session'
    api = 2

    def __init__(self):
        pass

    def apply(self, callback, route):
        @wraps(callback)
        def wrapper(*args, **kwargs):
            if self.name in route.skiplist:
                return callback(*args, **kwargs)
            data = callback(*args, **kwargs)
            if 'express.session' in request.environ:
                request.environ['express.session'].write()
                # session does not need to be close or released as
                # StrictRedis releases the connection after each operation
            return data
        return wrapper

    def setup(self, app):
        pass

    def close(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
from Cookie import SimpleCookie
from session import Session
from session import ExpressSession


class Dispatcher(object):

    """
    dispatch a request to one of the bundles controllers
    """

    def __init__(self, wsgi, config):
        self.app = wsgi
        self.config = config
        self.bundle = None
        self.controller = None
        if config.debug:
            self.imported = {}

    def route(self, path):
        segs = []
        if path != '/':
            segs = path.strip('/').split('/')
        if not len(segs):
            self.bundle = self.config.default['bundle']
            self.controller = self.config.default['controller']
        else:
            if self.is_bundle(segs[0]):
                self.bundle = segs[0]
                segs.pop(0)
                if not len(segs):
                    self.controller = self.config.default['controller']
                else:
                    if self.is_controller(self.bundle, segs[0]):
                        self.controller = segs[0]
                        segs.pop(0)
                    else:
                        self.controller = self.config.default['controller']
            elif self.is_bundle(self.config.default['bundle']):
                self.bundle = self.config.default['bundle']
                if self.is_controller(self.bundle, segs[0]):
                    self.controller = segs[0]
                    segs.pop(0)
                else:
                    self.controller = self.config.default['controller']

    def dispatch(self):
        def to_imported_key(name):
            return self.controller + '-' + name
        try:
            paths = self.config.apppath.strip(os.path.sep).split(os.path.sep)
            app = paths.pop()
            name = '.'.join(
                [app, 'bundles', self.bundle, 'controllers', self.controller])
            if name not in sys.modules:
                appdir = os.sep.join(path for path in paths)
                if appdir not in sys.path:
                    sys.path.insert(0, appdir)
                module = __import__(name, fromlist=[self.controller])
                if self.config.debug:
                    self.imported[to_imported_key(name)] = module
            elif self.config.debug:
                reload(self.imported[to_imported_key(name)])

        except Exception as e:
            import traceback
            tb = traceback.format_exc()
            print("Dispatcher: {0}".format(e))
            print(tb)

    def is_bundle(self, bundle):
        path = os.path.join(self.config.apppath, 'bundles', bundle)
        return os.path.isdir(path)

    def is_controller(self, bundle, controller):
        path = os.path.join(self.config.apppath, 'bundles',
                            bundle, 'controllers',
                            '.'.join([controller, 'py']))
        return os.path.isfile(path)

    def wsgi(self, environ, start_response):
        self.route(environ['PATH_INFO'])
        self.dispatch()
        return self.app

    def __call__(self, environ, start_response):
        return self.wsgi(environ, start_response)(environ, start_response)


class StripPath(object):

    def __init__(self, wsgi):
        self.app = wsgi

    def wsgi(self, environ, start_response):
        if environ['PATH_INFO'] != '/':
            environ['PATH_INFO'] = environ['PATH_INFO'].rstrip('/')
        return self.app

    def __call__(self, environ, start_response):
        return self.wsgi(environ, start_response)(environ, start_response)


class SessionWsgi(object):

    def __init__(self, wsgi):
        self.app = wsgi

    def wsgi(self, environ, start_response):
        return self.app

    def __call__(self, environ, start_response):
        sc = SimpleCookie()
        if 'HTTP_COOKIE' in environ:
            sc.load(environ.get('HTTP_COOKIE'))
        ses = Session(ExpressSession(sc))
        ses.read()
        environ['express.session'] = ses
        return self.app(environ, start_response)




class BaseError(Exception):

    def __init__(self, message):
        Exception.__init__(self, message)
        self.message = message


class ValidationError(BaseError):

    def __init__(self, message, fields={}):
        super(ValidationError, self).__init__(message)
        self.fields = fields


class ConnectionError(BaseError):

    def __init__(self, message, returned={}):
        super(ConnectionError, self).__init__(message)
        self.returned = returned



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from lib.decorator import get


@get('/', apply='json', skip='jinja2')
def index():
    return {'word': 'Hello Decanter!'}


@get('/home/')
@get('/home/<name>')
def home(name='Decanter'):
    return {'greeting': "Hello {0}!".format(name.title())}



"""
Import everything we want to have easily accessible to outside libraries,
but leave other initialization to decanter.py
"""

from lib.errors import BaseError, ValidationError, ConnectionError
from _version import __name__, __version__



import re
from gettext import gettext as _
from jsonschema import Draft4Validator


template_search = re.compile(r'{{[\s]?([\w]+)[\s]?}}')
template_replacement = '({{[\s]?%s[\s]?}})'


def _replace_vars(root, message):
    """
    Replace placeholders in a JSON-schema message in the {{var}} form
    with their values found in the JSON-schema definition.
    """
    while template_search.search(message):
        name = template_search.search(message).groups()[0]
        message = re.sub(template_replacement %
                         name, root.get(name, "this"), message)
    return message.capitalize()


def get_error_dictionary(schema, instance):
    """
    Start with jsonschema and instance to test against, and return
    a dictionary with field names as keys and error messages as values,
    to be used directly by the front-end.
    """

    v = Draft4Validator(schema)
    errors = v.iter_errors(instance)
    d = {}

    # iterate through the validation errors
    for error in errors:
        root = error.schema
        # todo: add $ref handling
        if error.schema_path[0] == 'properties':
            # traverse the tree
            for i in range(1, len(error.schema_path) - 1):
                root = root.get(error.schema_path[i], root)

            # read the error messages
            if "errors" in root:
                message = root.get("errors", {}).get(error.schema_path[-1])
                d[error.schema_path[1]] = _replace_vars(root, message)
            else:
                # should never happen
                d[error.schema_path[1]] = _("Error when validating")
        else:
            # special case when error is not on a specific property
            message = root.get("errors", {}).get(error.schema_path[0])
            try:
                d[error.schema_path[1]] = _replace_vars(root, message)
            except IndexError:
                d[error.schema_path[0]] = _replace_vars(root, message)

    return d
